# container-service-extension
# Copyright (c) 2017 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause


class VC_Adapter(object):

    def __init__(self, config, vca_system, prov):
        self.config = config
        self.vca_system = vca_system
        self.prov = prov

    def get_list_params(self):
        """
        Translates the input parameters identifying a list cluster operation in
        vCD/CSE to the input parameters required by KoV to list clusters.
        It uses the global config, self.vca_system and self.prov to
        introspect vCD.

        :return: (result)
        """

        vcs = self.config['vcs'][0]

        return vcs

    def get_create_params(self, body, cluster_id):
        """
        Translates the input parameters identifying a cluster in vCD to
        the input parameters required by KoV to create a cluster.
        It uses self.vca_system and self.prov to introspect vCD.
        The result should also include the vCenter information.

        :return: (result)
        """

        # vdc_name = body['vdc']
        # request_username = self.prov.vca_tenant.vcloud_session.username
        # request_org_id = self.prov.vca_tenant.vcloud_session.org_id

        vcs = self.config['vcs'][0]

        params = {
                    "name": 'c-' + cluster_id,
                    "cloudProvider": "vsphere",
                    "datacenter": vcs['datacenter'],
                    "datastore": vcs['datastore'],
                    "maxNodes": body['node_count'],
                    "minNodes": body['node_count'],
                    "noOfMasters": 1,
                    "nodeNetwork": vcs['network'],
                    "managementNetwork": vcs['network'],
                    "opsUsername": vcs['username'],
                    "opsPassword": vcs['password'],
                    "vsphereCluster": vcs['cluster'],
                    "authorizedKeys": []
                 }

        # params = {
        #             "name": 'c-' + cluster_id,
        #             "minNodes": body['node_count'],
        #             "maxNodes": body['node_count'],
        #             "authorizedKeys": [],
        #             "vsphere": {
        #                 "datacenter": vcs['datacenter'],
        #                 "computeResource": vcs['cluster'],
        #                 "datastore": vcs['datastore'],
        #                 "publicNetwork": vcs['network'],
        #                 "opsUsername": vcs['username'],
        #                 "opsPassword": vcs['password'],
        #             }
        #          }

        return (vcs, params)

    def get_delete_params(self, body, cluster_id):
        """
        Translates the input parameters identifying a cluster in vCD to
        the input parameters required by KoV to delete a cluster.
        It uses self.vca_system and self.prov to introspect vCD.
        The result should also include the vCenter information.

        :return: (result)
        """

        vcs = self.config['vcs'][0]
        return(vcs, 'c-'+cluster_id)



# container-service-extension
# Copyright (c) 2017 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

import base64
import json
import jsonpickle
import logging
import pika
from processor import ServiceProcessor
import thread
import traceback

LOGGER = logging.getLogger(__name__)


class MessageConsumer(object):
    EXCHANGE_TYPE = 'direct'

    def __init__(self, amqp_url, exchange, routing_key,
                 config, verify, log=False):
        self._connection = None
        self._channel = None
        self._closing = False
        self._consumer_tag = None
        self._url = amqp_url
        self.exchange = exchange
        self.routing_key = routing_key
        self.queue = routing_key
        self.config = config
        self.verify = verify
        self.log = log

    def connect(self):
        LOGGER.info('Connecting to %s', self._url)
        return pika.SelectConnection(pika.URLParameters(self._url),
                                     self.on_connection_open,
                                     stop_ioloop_on_close=False)

    def on_connection_open(self, unused_connection):
        LOGGER.debug('Connection opened')
        self.add_on_connection_close_callback()
        self.open_channel()

    def add_on_connection_close_callback(self):
        LOGGER.debug('Adding connection close callback')
        self._connection.add_on_close_callback(self.on_connection_closed)

    def on_connection_closed(self, connection, reply_code, reply_text):
        self._channel = None
        if self._closing:
            self._connection.ioloop.stop()
        else:
            LOGGER.warning('Connection closed, reopening in 5 seconds: '
                           '(%s) %s', reply_code, reply_text)
            self._connection.add_timeout(5, self.reconnect)

    def reconnect(self):
        self._connection.ioloop.stop()

        if not self._closing:
            self._connection = self.connect()
            self._connection.ioloop.start()

    def open_channel(self):
        LOGGER.debug('Creating a new channel')
        self._connection.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel):
        LOGGER.debug('Channel opened')
        self._channel = channel
        self.add_on_channel_close_callback()
        self.setup_exchange(self.exchange)

    def add_on_channel_close_callback(self):
        LOGGER.debug('Adding channel close callback')
        self._channel.add_on_close_callback(self.on_channel_closed)

    def on_channel_closed(self, channel, reply_code, reply_text):
        LOGGER.warning('Channel %i was closed: (%s) %s',
                       channel, reply_code, reply_text)
        self._connection.close()

    def setup_exchange(self, exchange_name):
        LOGGER.debug('Declaring exchange %s', exchange_name)
        self._channel.exchange_declare(self.on_exchange_declareok,
                                       exchange=exchange_name,
                                       type=self.EXCHANGE_TYPE,
                                       durable=True)

    def on_exchange_declareok(self, unused_frame):
        LOGGER.debug('Exchange declared')
        self.setup_queue(self.queue)

    def setup_queue(self, queue_name):
        LOGGER.debug('Declaring queue %s', queue_name)
        self._channel.queue_declare(self.on_queue_declareok, queue_name)

    def on_queue_declareok(self, method_frame):
        LOGGER.debug('Binding %s to %s with %s',
                     self.exchange, self.queue, self.routing_key)
        self._channel.queue_bind(self.on_bindok, self.queue,
                                 self.exchange, self.routing_key)

    def on_bindok(self, unused_frame):
        LOGGER.debug('Queue bound')
        self.start_consuming()

    def start_consuming(self):
        LOGGER.debug('Issuing consumer related RPC commands')
        self.add_on_cancel_callback()
        self._consumer_tag = self._channel.basic_consume(self.on_message,
                                                         self.queue)

    def add_on_cancel_callback(self):
        LOGGER.debug('Adding consumer cancellation callback')
        self._channel.add_on_cancel_callback(self.on_consumer_cancelled)

    def on_consumer_cancelled(self, method_frame):
        LOGGER.debug('Consumer was cancelled remotely, shutting down: %r',
                     method_frame)
        if self._channel:
            self._channel.close()

    def on_message(self, unused_channel, basic_deliver, properties, body):
        LOGGER.debug('Received message # %s from %s (%s): %s',
                     basic_deliver.delivery_tag,
                     properties.app_id,
                     thread.get_ident(),
                     json.dumps(json.loads(body)[0]))
        self.acknowledge_message(basic_deliver.delivery_tag)
        try:
            sp = ServiceProcessor(self.config,
                                  self.verify,
                                  self.log)
            body = json.loads(body)[0]
            result = sp.process_request(body)
            reply_body = jsonpickle.encode(result['body'])
            status_code = result['status_code']
        except:
            reply_body = '{}'
            status_code = 500
            tb = traceback.format_exc()
            LOGGER.error(tb)
        reply_msg = {
            'id': body['id'],
            'headers': {'Content-Type': body['headers']['Accept'],
                        'Content-Length': len(reply_body)},
            'statusCode': status_code,
            'body': base64.b64encode(reply_body),
            'request': False
        }
        LOGGER.debug('reply: %s', json.dumps(reply_body))

        reply_properties = pika.BasicProperties(
                               correlation_id=properties.correlation_id)
        result = self._channel.basic_publish(
            exchange=properties.headers['replyToExchange'],
            routing_key=properties.reply_to,
            body=json.dumps(reply_msg),
            properties=reply_properties)

    def acknowledge_message(self, delivery_tag):
        LOGGER.debug('Acknowledging message %s', delivery_tag)
        self._channel.basic_ack(delivery_tag)

    def stop_consuming(self):
        if self._channel:
            LOGGER.info('Sending a Basic.Cancel RPC command to RabbitMQ')
            self._channel.basic_cancel(self.on_cancelok, self._consumer_tag)

    def on_cancelok(self, unused_frame):
        LOGGER.debug('RabbitMQ acknowledged the cancellation'
                     'of the consumer')
        self.close_channel()

    def close_channel(self):
        LOGGER.debug('Closing the channel')
        self._channel.close()

    def run(self):
        self._connection = self.connect()
        self._connection.ioloop.start()

    def stop(self):
        LOGGER.info('Stopping')
        self._closing = True
        self.stop_consuming()
        self._connection.ioloop.start()
        LOGGER.info('Stopped')

    def close_connection(self):
        LOGGER.info('Closing connection')
        self._connection.close()



# container-service-extension
# Copyright (c) 2017 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

import logging
import time
from pyvcloud.task import Task
from pyvcloud.vcloudair import VCA
from pyvcloud.vcloudsession import VCS


LOGGER = logging.getLogger(__name__)


def wait_for_task(task, task_id):
    t = task.get_task(task_id)
    while t.get_status() != 'success':
        time.sleep(3)
        t = task.get_task(task_id)
        LOGGER.debug('%s, %s, %s, %s, %s, %s->%s' %
                     (t.get_id().split(':')[-1],
                      t.get_operation(),
                      t.get_Owner().get_name(),
                      t.get_status(),
                      t.get_Progress(),
                      str(t.get_startTime()).split('.')[0],
                      str(t.get_endTime()).split('.')[0]))


class Provisioner(object):

    def __init__(self, host, vcloud_token, version, verify, log=False):
        self.host = host
        self.vcloud_token = vcloud_token
        self.version = version
        self.verify = verify
        self.log = log
        self.vca_tenant = VCA(host=host, username='', service_type='vcd',
                              version=version, verify=verify, log=log)

    def connect(self):
        link = 'https://%s/api/session' % (self.host)
        vcloud_session = VCS(link, '', '', None, '', '',
                             version=self.version,
                             verify=self.verify,
                             log=self.log)
        result = vcloud_session.update_session_data(self.vcloud_token)
        if not result:
            LOGGER.error('unable to connect provisioner for token: %s' %
                         (self.vcloud_token))
            return False
        self.vca_tenant.vcloud_session = vcloud_session
        LOGGER.info('connected provisioner for: %s@%s (%s)' %
                    (self.vca_tenant.vcloud_session.username,
                     self.vca_tenant.vcloud_session.org,
                     self.vcloud_token))
        return True

    def get_task_status(self, task_id):
        try:
            task = Task(session=self.vca_tenant.vcloud_session,
                        verify=self.verify, log=self.log)
            t = task.get_task(task_id)
            return t.get_status()
        except:
            return 'unknown'

    def validate_name(self, name):
        """
        Validates that the cluster name against the pattern.
        """

        # TODO (validate against pattern)
        # pattern = '^[a-zA-Z](([-0-9a-zA-Z]+)?[0-9a-zA-Z])?(\.[a-zA-Z](([-0-9a-zA-Z]+)?[0-9a-zA-Z])?)*$'  # NOQA

        return True

    def search_by_name(self, name):
        """
        check that the cluster name exists in the current VDC.
        It exists, it returns the cluster id
        """
        result = {'name': name, 'cluster_id': None}

        return result

    def search_by_id(self, cluster_id):
        """
        check that the cluster with cluster_id exists in the current VDC.
        It exists, it returns the cluster name and details.
        """
        # result = {'name': None, 'cluster_id': cluster_id}
        result = {'name': cluster_id.split('-')[0], 'cluster_id': cluster_id}

        return result

    @staticmethod
    def get_system_session(config):
        vca_system = VCA(host=config['vcd']['host'],
                         username=config['vcd']['username'],
                         service_type='vcd',
                         version=config['vcd']['api_version'],
                         verify=config['vcd']['verify'],
                         log=config['vcd']['log'])

        org_url = 'https://%s/cloud' % config['vcd']['host']
        r = vca_system.login(password=config['vcd']['password'],
                             org='System',
                             org_url=org_url)
        if not r:
            raise Exception('failed to login as system')
        r = vca_system.login(token=vca_system.token,
                             org='System',
                             org_url=vca_system.vcloud_session.org_url)
        if not r:
            raise Exception('failed to login as system')
        return vca_system



#!/usr/bin/env python

# container-service-extension
# Copyright (c) 2017 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

from consumer import MessageConsumer
import logging
import os
import pkg_resources
import signal
import sys
from threading import Thread
import time
import traceback
import yaml
import pika
from pyvcloud.vcloudair import VCA
from utils import get_thumbprint


LOGGER = logging.getLogger(__name__)
config = {}


def print_help():
    print('Container Service Extension for vCloud Director, version %s'
          % pkg_resources.require("container-service-extension")[0].version)
    print("""
Usage: cse COMMAND [ARGS]

Commands:
  help                  Prints this messages
  version               Shows version
  init [config.yml]     Creates config.yml file with default values
  check <config.yml>    Checks configuration in config.yml (AMQP and vCD)
  run <config.yml>      Run cse with options from config.yml
""")
    sys.exit(0)


def init(file_name='config.yml'):
    default_config = """
amqp:
    host: amqp.vmware.com
    port: 5672
    user: 'guest'
    password: 'guest'
    exchange: vcdext
    routing_key: cse

vcd:
    host: vcd.vmware.com
    port: 443
    username: 'administrator'
    password: 'my_secret_password'
    api_version: '6.0'
    verify: False
    log: True

vcs:
  - host: vcs.vmware.com
    port: 443
    username: 'administrator@vsphere.local'
    password: 'my_secret_password'
    verify: False
    datacenter: 'my_datacenter'
    cluster: 'my_cluster'
    datastore: 'my_datastore'
    network: 'my_network'

service:
    listeners: 2
    logging_level: 5
    logging_format: %s
    key_filename: 'id_rsa_cse'
    key_filename_pub: 'id_rsa_cse.pub'

    """ % '%(levelname) -8s %(asctime)s %(name) -40s %(funcName) ' \
          '-35s %(lineno) -5d: %(message)s'
    if os.path.isfile(file_name):
        print('file %s already exist, aborting' % file_name)
        sys.exit(1)
    with open(file_name, 'w') as f:
        f.write(default_config)
    print('default config saved to file \'%s\'' % file_name)


def bool_to_msg(value):
    # if value:
    #     return u'\u2714'
    # else:
    #     return u'\u2718'
    if value:
        return 'success'
    else:
        return 'fail'


def check_config(file_name):
    config = {}
    try:
        with open(file_name, 'r') as f:
            config = yaml.load(f)
    except:
        print('config file \'%s\' not found or invalid' % file_name)
        sys.exit(1)
    try:
        amqp = config['amqp']
        credentials = pika.PlainCredentials(amqp['user'], amqp['password'])
        parameters = pika.ConnectionParameters(amqp['host'], amqp['port'],
                                               '/',
                                               credentials)
        connection = pika.BlockingConnection(parameters)
        print('Connection to AMQP server (%s:%s): %s' % (amqp['host'],
              amqp['port'],
              bool_to_msg(connection.is_open)))
        connection.close()
        vca_system = VCA(host=config['vcd']['host'],
                         username=config['vcd']['username'],
                         service_type='standalone',
                         version=config['vcd']['api_version'],
                         verify=config['vcd']['verify'],
                         log=config['vcd']['log'])

        org_url = 'https://%s/cloud' % config['vcd']['host']
        r = vca_system.login(password=config['vcd']['password'],
                             org='System',
                             org_url=org_url)
        print('Connection to vCloud Director (%s:%s): %s' %
              (config['vcd']['host'], config['vcd']['port'],
               bool_to_msg(r)))
        if r:
            r = vca_system.login(token=vca_system.token,
                                 org='System',
                                 org_url=vca_system.vcloud_session.org_url)
            print('  login to \'System\' org: %s' % (bool_to_msg(r)))

        for vc in config['vcs']:
            r = get_thumbprint(vc['host'], vc['port'])
            print('Connection to vCenter Server (%s:%s): %s' %
                  (vc['host'], vc['port'],
                   bool_to_msg(len(r) > 0)))
    except:
        tb = traceback.format_exc()
        print('failed to validate configuration from file %s' % file_name)
        print(tb)
        sys.exit(1)


def signal_handler(signal, frame):
    print('\nCrtl+C detected, exiting')
    raise KeyboardInterrupt()


def consumer_thread(c):
    try:
        LOGGER.info('about to start consumer_thread %s', c)
        c.run()
    except:
        print('about to stop consumer_thread')
        print(traceback.format_exc())
        c.stop()


def main():
    global config
    import requests
    requests.packages.urllib3.disable_warnings()
    if len(sys.argv) > 0:
        if len(sys.argv) > 1:
            if sys.argv[1] == 'version':
                version = pkg_resources.\
                          require("container-service-extension")[0].version
                print(version)
                sys.exit(0)
            elif sys.argv[1] == 'help':
                print_help()
                sys.exit(0)
            elif sys.argv[1] == 'init':
                if len(sys.argv) > 2:
                    init(sys.argv[2])
                else:
                    init()
                sys.exit(0)
            elif sys.argv[1] == 'check':
                check_config(sys.argv[2])
                sys.exit(0)
            elif sys.argv[1] == 'run':
                pass
        else:
            print_help()
    try:
        with open(sys.argv[2], 'r') as f:
            config = yaml.load(f)
    except:
        tb = traceback.format_exc()
        print('config file \'%s\' not found or invalid' % sys.argv[1])
        print(tb)
        sys.exit(1)

    logging.basicConfig(filename='cse.log',
                        level=config['service']['logging_level'],
                        format=config['service']['logging_format'])

    print('Container Service Extension for vCloud Director running')
    print('see file \'cse.log\' for details')
    print('press Ctrl+C to finish')
    signal.signal(signal.SIGINT, signal_handler)

    LOGGER.info('Container Service Extension for vCloud Director')
    LOGGER.info('waiting for requests...')

    amqp = config['amqp']
    num_consumers = config['service']['listeners']
    consumers = []
    threads = []

    for n in range(num_consumers):
        try:
            c = MessageConsumer('amqp://%s:%s@%s:%s/' %
                                (amqp['user'],
                                 amqp['password'],
                                 amqp['host'],
                                 amqp['port']),
                                amqp['exchange'],
                                amqp['routing_key'],
                                config,
                                config['vcd']['verify'],
                                config['vcd']['log'])
            t = Thread(target=consumer_thread, args=(c,))
            t.daemon = True
            t.start()
            LOGGER.info('started thread %s', t.ident)
            threads.append(t)
            consumers.append(c)
            time.sleep(0.25)
        except KeyboardInterrupt:
            break
        except:
            print(traceback.format_exc())

    LOGGER.info('num of threads started: %s', len(threads))

    while True:
        try:
            time.sleep(1)
        except KeyboardInterrupt:
            LOGGER.info('interrupt detected')
            LOGGER.info('closing connections...')
            for c in consumers:
                try:
                    c.stop()
                except:
                    pass
            LOGGER.info('done')
            break
        except:
            print(traceback.format_exc())
            sys.exit(1)


if __name__ == '__main__':
    main()



# container-service-extension
# Copyright (c) 2017 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

import json
import logging

TYPE_MASTER = 'master'
TYPE_NODE = 'node'
LOGGER = logging.getLogger(__name__)


class Node(object):

    def __init__(self, name, node_type=TYPE_NODE, node_id='', href=''):
        self.name = name
        self.node_type = node_type
        self.node_id = node_id
        self.href = href
        self.ip = ''
        self.cluster_id = ''
        self.cluster_name = ''

    def __repr__(self):
        return self.toJSON()

    def toJSON(self):
        return json.dumps(self,
                          default=lambda o: o.__dict__,
                          sort_keys=True,
                          indent=4)


class Cluster(object):

    def __init__(self, name=None, cluster_id=''):
        self.name = name
        self.cluster_id = cluster_id
        self.master_nodes = []
        self.nodes = []
        self.vdc = None
        self.status = None
        self.leader_endpoint = None

    def __repr__(self):
        return self.toJSON()

    def toJSON(self):
        return json.dumps(self,
                          default=lambda o: o.__dict__,
                          sort_keys=True,
                          indent=4)
